package com.serialio.sdm_tcpip_client.presentation.main

import androidx.lifecycle.LiveData
import androidx.lifecycle.MutableLiveData
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import com.serialio.sdm_tcpip_client.App
import com.serialio.sdm_tcpip_client.data.repository.sdm_device.model.SdmDevice
import com.serialio.sdm_tcpip_client.data.tcp_protocol.ConstantsTcp
import com.serialio.sdm_tcpip_client.data.tcp_protocol.isJSONValid
import com.serialio.sdm_tcpip_client.data.tcp_protocol.makeClientIdentificationRequest
import com.serialio.sdm_tcpip_client.data.tcp_protocol.makeDeviceCommandRequest
import com.serialio.sdm_tcpip_client.data.tcp_protocol.makeDeviceConnectRequest
import com.serialio.sdm_tcpip_client.data.tcp_protocol.makeRemoteDeviceListRequest
import com.serialio.sdm_tcpip_client.data.tcp_protocol.parseServerJsonData
import com.serialio.sdm_tcpip_client.data.upd_client.data.ConstantsUdp
import com.serialio.sdm_tcpip_client.data.upd_client.data.DataUdp
import com.serialio.sdm_tcpip_client.domain.ServerResponseEngine
import com.serialio.sdm_tcpip_client.domain.UdpEngine
import com.serialio.sdm_tcpip_client.domain.repository.TcpClientRepository
import com.serialio.sdm_tcpip_client.utils.getIpv4HostAddress
import com.serialio.sdm_tcpip_client.utils.ClientConnectionStatus
import com.serialio.sdm_tcpip_client.utils.PeripheralConnectionStatus
import com.serialio.sdm_tcpip_client.utils.ServerResponse
import dagger.hilt.android.lifecycle.HiltViewModel
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.Job
import kotlinx.coroutines.flow.cancellable
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import javax.inject.Inject

@HiltViewModel
class MainActivityViewModel @Inject constructor(
    private val tcpClientRepository: TcpClientRepository,
    private val serverDataEngine: ServerResponseEngine,
    private val udpEngine: UdpEngine,
) : ViewModel() {
    var requestIdInctement = 1
    lateinit var clientID: String
    var job: Job? = null

    private val _serverConnectionStatusLiveData = MutableLiveData<ClientConnectionStatus>()
    val serverConnectionStatusLiveData: LiveData<ClientConnectionStatus> =
        _serverConnectionStatusLiveData

    private val _sdmDevicesLiveData = MutableLiveData<List<SdmDevice>>()
    val sdmDevicesLiveData: LiveData<List<SdmDevice>> = _sdmDevicesLiveData

    private val _serverResponseLiveData = MutableLiveData<ServerResponse>()
    val serverResponseLiveData: LiveData<ServerResponse> = _serverResponseLiveData

    private val _dataSentToServerLiveData = MutableLiveData<String>()
    val dataSentToServerLiveData: LiveData<String> = _dataSentToServerLiveData

    private val _sdmSelectedDevice = MutableLiveData<SdmDevice>()
    val sdmSelectedDevice: LiveData<SdmDevice> = _sdmSelectedDevice

    private val _updData = MutableLiveData<DataUdp>()
    val updData: LiveData<DataUdp> = _updData

    private val _serverIP = MutableLiveData<String>()
    val serverIP: LiveData<String> = _serverIP

    private val _peripheralConnectionStatusDeviceLiveData =
        MutableLiveData<PeripheralConnectionStatus>()
    val peripheralConnectionStatusDeviceLiveData: LiveData<PeripheralConnectionStatus> =
        _peripheralConnectionStatusDeviceLiveData

    init {
        getSdmDeices()

        viewModelScope.launch() {
            serverDataEngine.serverResponseFlow.collect { serverResponse ->
                _serverResponseLiveData.value = serverResponse
                when (serverResponse) {
                    is ServerResponse.ClientIdentification -> {
                        if (serverResponse.isSuccess) {

                            // Send DeviceList Request to SioSdmServer (If identification is approved)
                            requestIdInctement++
                            val sRequestDeviceList = makeRemoteDeviceListRequest(
                                request_id = requestIdInctement,
                                client_id = clientID
                            )
                            tcpClientRepository.sendMsg(sRequestDeviceList)
                        }
                    }

                    is ServerResponse.PeripheralDeviceList -> {
                        if (serverResponse.isSuccess && !serverResponse.params?.remote_devices.isNullOrEmpty()) {

                            // Clear current device list
                            serverDataEngine.sdmDeviceRepository.clearDevices()

                            // Save new device list from SioSdmServer
                            serverResponse.params?.remote_devices?.map { it.toSdmDevice() }
                                ?.let { serverDataEngine.sdmDeviceRepository.addDevices(it) }

                            // Send Selected device status request (if we have selected desire device)
                            if (_sdmSelectedDevice.value != null &&
                                !_sdmSelectedDevice.value!!.deviceId.isNullOrEmpty()
                            ) {
                                val findDevice =
                                    serverResponse.params?.remote_devices?.find { it.device_id == _sdmSelectedDevice.value?.deviceId }
                                if (findDevice != null)
                                    sendConnectDeviceRequest(_sdmSelectedDevice.value!!.deviceId)
                            }
                        }
                    }

                    is ServerResponse.PeripheralDeviceStatus -> {
                        if (_sdmSelectedDevice.value != null) {
                            var status = ConstantsTcp.DEVICE_STATE_UNAVAILABLE
                            if (serverResponse.isSuccess) {
                                status = serverResponse.params?.device_state.toString()
                            }

                            if (serverResponse.params?.device_id == _sdmSelectedDevice.value!!.deviceId &&
                                _sdmSelectedDevice.value!!.status != status
                            ) {
                                App.logger.putt("Update device status from request:\n deviceId:${_sdmSelectedDevice.value!!.deviceId}, status: ${status}\n")
                                _sdmSelectedDevice.value!!.status = status

                                serverDataEngine.sdmDeviceRepository.updateStatus(
                                    _sdmSelectedDevice.value!!.deviceId,
                                    status
                                )
                            }
                        }
                    }

                    is ServerResponse.PeripheralDeviceCmdResponse -> {

                    }
                }
            }
        }
        viewModelScope.launch() {
            serverDataEngine.peripheralConnectionStatusFlow.collect { peripheralConnectioStatus ->
                _peripheralConnectionStatusDeviceLiveData.value = peripheralConnectioStatus
                if (peripheralConnectioStatus.deviceID != null && peripheralConnectioStatus.status != null) {

                    serverDataEngine.sdmDeviceRepository.updateStatus(
                        peripheralConnectioStatus.deviceID,
                        peripheralConnectioStatus.status
                    )
                }
            }
        }

        viewModelScope.launch() {
            tcpClientRepository.dataSentToServer().collect { dataSent ->
                _dataSentToServerLiveData.value = dataSent
            }
        }

        viewModelScope.launch() {
            startUPD(
                ConstantsUdp.UDP_CLIENT_PORT,
                ConstantsUdp.UDP_SERVER_PORT,
                getIpv4HostAddress()
            )
        }
    }

    fun sendCmd(device_id: String, command_name: String) {
        viewModelScope.launch {
            requestIdInctement++
            tcpClientRepository.sendMsg(
                makeDeviceCommandRequest(
                    request_id = requestIdInctement,
                    client_id = clientID,
                    device_id = device_id,
                    command_name = command_name
                )
            )
        }
    }

    fun startUPD(
        portLocal: Int,
        portRemote: Int,
        ipLocal: String,
    ) {
        viewModelScope.launch {
            udpEngine.listenUdpData(portLocal, portRemote, ipLocal).collect {
                if (!it.ip.isEmpty() && it.message.startsWith(ConstantsUdp.UDP_SERVER_RESPONSE)) {
                    _serverIP.value = it.ip
                    udpEngine.stopTimer()
                    udpEngine.stopUpdListening()
                }
                _updData.value = it
            }
        }

        viewModelScope.launch {
            udpEngine.startUdpRequestsTimer(ipLocal)
        }

    }

    private fun getSdmDeices() {
        App.logger.putt("MainActivityViewModel.getSdmDeices\n")
        viewModelScope.launch() {
            serverDataEngine.sdmDeviceRepository.getDevices().collect {
                App.logger.putt("MainActivityViewModel.sdmDeviceRepository.SDM Devices ${it.size}\n")
                _sdmDevicesLiveData.value = it

                val findDevice = it.find { it.deviceId == _sdmSelectedDevice.value?.deviceId }
                findDevice?.let {
                    _sdmSelectedDevice.value = findDevice!!
                }
            }
        }
    }

    fun disconnectServer() {
        App.logger.putt("MainActivityViewModel.disconnectServer\n")
        viewModelScope.launch() {
            tcpClientRepository.stopClient()
        }
    }

    fun connectServer(ip: String, port: Int) {
        App.logger.putt("MainActivityViewModel.connectServer: ${ip}:${port}\n")
        job?.cancel()
        job = viewModelScope.launch() {
            tcpClientRepository.connect(ip, port).cancellable().collect { connectionStatus ->
                withContext(Dispatchers.Main) {
                    _serverConnectionStatusLiveData.value = connectionStatus
                    when (connectionStatus) {
                        is ClientConnectionStatus.OnDataReceived -> {
                            connectionStatus.message?.let {
                                if (isJSONValid(connectionStatus.message)) {
                                    val parsed = parseServerJsonData(connectionStatus.message)
                                    parsed?.let {
                                        serverDataEngine.parseServerData(parsed)
                                    }
                                } else {
                                    App.logger.putt("isNotJSON data:\n${connectionStatus.message}\n")
                                }
                            }
                        }

                        is ClientConnectionStatus.Connected -> {
                            // After connect we send Client Identification  request
                            requestIdInctement++
                            sendMsg(
                                makeClientIdentificationRequest(
                                    requestIdInctement,
                                    clientID,
                                    "serialio"
                                )
                            )
                        }

                        is ClientConnectionStatus.Disconnected -> {
                            if (_sdmSelectedDevice.value != null)
                                serverDataEngine.sdmDeviceRepository.updateStatus(
                                    _sdmSelectedDevice.value!!.deviceId,
                                    ""
                                )
                        }

                        ClientConnectionStatus.Connecting -> {

                        }

                        is ClientConnectionStatus.Error -> {

                        }
                    }
                }
            }
        }
    }

    fun sendMsg(msg: String) {
        App.logger.putt("MainActivityViewModel.sendMsg: ${msg}\n")
        viewModelScope.launch() {
            tcpClientRepository.sendMsg(msg)
        }
    }

    fun sendCmdToSelectedDevice(sCmd: String) {
        App.logger.putt("MainActivityViewModel.sendCmdToSelectedDevice: ${sCmd}\n")
        val device = sdmSelectedDevice
        if (device.value != null) {
            if (device.value?.status == ConstantsTcp.DEVICE_STATE_READY || device.value?.status == ConstantsTcp.DEVICE_STATE_CONNECTED) {
                requestIdInctement++
                val msg =
                    makeDeviceCommandRequest(
                        requestIdInctement,
                        clientID,
                        device.value!!.deviceId,
                        sCmd
                    )
                viewModelScope.launch() {
                    tcpClientRepository.sendMsg(msg)
                }
                _peripheralConnectionStatusDeviceLiveData.value =
                    PeripheralConnectionStatus(status = "Command Sent");
            } else
                _peripheralConnectionStatusDeviceLiveData.value =
                    PeripheralConnectionStatus(status = "Device not connected!");
        } else {
            _peripheralConnectionStatusDeviceLiveData.value =
                PeripheralConnectionStatus(status = "No selected device!");
        }
    }

    fun updateSelectedDevice(sdmDevice: SdmDevice) {
        App.logger.putt("updateSelectedDevice:\n${sdmDevice}\n")
        _sdmSelectedDevice.value = sdmDevice
    }

    fun sendConnectDeviceRequest(deviceId: String?) {
        App.logger.putt("MainActivityViewModel.sendConnectDeviceRequest: ${deviceId}\n")
        if (deviceId != null) {
            requestIdInctement++
            val sRequestDeviceList = makeDeviceConnectRequest(
                request_id = requestIdInctement,
                client_id = clientID,
                device_id = deviceId,
                connect_disconnect = ConstantsTcp.REQUEST_CONNECT
            )
            viewModelScope.launch() {
                tcpClientRepository.sendMsg(sRequestDeviceList)
            }
        }
    }
}
